import pandas as pd

import dtale.global_state as global_state
from dtale.utils import make_list, run_query


def flatten_columns(df, columns=None):
    if columns is not None:
        return [
            ' '.join(['{}-{}'.format(c1, str(c2)) for c1, c2 in zip(make_list(columns), make_list(col_val))]).strip()
            for col_val in df.columns.values
        ]
    return [' '.join([str(c) for c in make_list(col)]).strip() for col in df.columns.values]


class DataReshaper(object):

    def __init__(self, data_id, shape_type, cfg):
        self.data_id = data_id
        if shape_type == 'pivot':
            self.builder = PivotBuilder(cfg)
        elif shape_type == 'aggregate':
            self.builder = AggregateBuilder(cfg)
        elif shape_type == 'transpose':
            self.builder = TransposeBuilder(cfg)
        else:
            raise NotImplementedError('{} data re-shaper not implemented yet!'.format(shape_type))

    def reshape(self):
        data = run_query(
            global_state.get_data(self.data_id),
            (global_state.get_settings(self.data_id) or {}).get('query'),
            global_state.get_context_variables(self.data_id)
        )
        return self.builder.reshape(data)

    def build_code(self):
        return self.builder.build_code()


class PivotBuilder(object):

    def __init__(self, cfg):
        self.cfg = cfg

    def reshape(self, data):
        index, columns, values, aggfunc = (self.cfg.get(p) for p in ['index', 'columns', 'values', 'aggfunc'])
        if aggfunc is not None or len(values) > 1:
            pivot_data = pd.pivot_table(data, values=values, index=index, columns=columns, aggfunc=aggfunc)
            if len(values) == 1:
                pivot_data.columns = pivot_data.columns.droplevel(0)
        else:
            pivot_data = data.pivot(index=index, columns=columns, values=values[0])
        if self.cfg.get('columnNameHeaders', False):
            pivot_data.columns = flatten_columns(pivot_data, columns=columns)
        else:
            pivot_data.columns = flatten_columns(pivot_data)
        pivot_data = pivot_data.rename_axis(None, axis=1)
        return pivot_data

    def build_code(self):
        index, columns, values, aggfunc = (self.cfg.get(p) for p in ['index', 'columns', 'values', 'aggfunc'])
        code = []
        if aggfunc is not None or len(values) > 1:
            code.append("df = pd.pivot_table(df, index='{}', columns='{}', values=['{}'], aggfunc='{}')".format(
                index, columns, "', '".join(values), aggfunc
            ))
            if len(values) > 1:
                code.append(
                    "df.columns = [' '.join([str(c) for c in col]).strip() for col in df.columns.values]"
                )
            elif len(values) == 1:
                code.append("df.columns = df.columns.droplevel(0)")
        else:
            code.append("df = df.pivot(index='{index}', columns='{columns}', values='{values}')".format(
                index=index, columns=columns, values=values[0]
            ))
        code.append('df = df.rename_axis(None, axis=1)')
        return '\n'.join(code)


class AggregateBuilder(object):

    def __init__(self, cfg):
        self.cfg = cfg

    def reshape(self, data):
        index, agg = (self.cfg.get(p) for p in ['index', 'agg'])
        agg_data = data.groupby(index)
        agg_type, func, cols = (agg.get(p) for p in ['type', 'func', 'cols'])
        if agg_type == 'func':
            if cols:
                agg_data = agg_data[cols]
            return getattr(agg_data, func)()
        agg_data = agg_data.aggregate(cols)
        agg_data.columns = flatten_columns(agg_data)
        return agg_data

    def build_code(self):
        index, agg = (self.cfg.get(p) for p in ['index', 'agg'])
        index = "', '".join(index)
        agg_type, func, cols = (agg.get(p) for p in ['type', 'func', 'cols'])
        if agg_type == 'func':
            if cols is not None:
                return "df = df.groupby(['{index}'])['{columns}'].{agg}()".format(
                    index=index, columns="', '".join(cols), agg=agg
                )
            return "df = df.groupby(['{index}']).{agg}()".format(index="', '".join(index), agg=agg)
        code = [
            "df = df.groupby(['{index}']).aggregate(".format(index=index) + "{",
            ',\n'.join("\t'{col}': ['{aggs}']".format(col=col, aggs="', '".join(aggs)) for col, aggs in cols.items()),
            "})",
            "df.columns = [' '.join([str(c) for c in col]).strip() for col in df.columns.values]"
        ]
        return '\n'.join(code)


class TransposeBuilder(object):

    def __init__(self, cfg):
        self.cfg = cfg

    def reshape(self, data):
        index, columns = (self.cfg.get(p) for p in ['index', 'columns'])
        t_data = data.set_index(index)
        if any(t_data.index.duplicated()):
            raise Exception('Transposed data contains duplicates, please specify additional index or filtering')
        if columns is not None:
            t_data = t_data[columns]
        t_data = t_data.T
        if len(index) > 1:
            t_data.columns = flatten_columns(t_data)
        t_data = t_data.rename_axis(None, axis=1)
        return t_data

    def build_code(self):
        index, columns = (self.cfg.get(p) for p in ['index', 'columns'])

        code = []
        if columns is not None:
            code.append("df = df.set_index('{}')['{}'].T".format("', '".join(index), "', '".join(columns)))
        else:
            code.append("df = df.set_index('{}').T".format("', '".join(index)))
        if len(index) > 1:
            code.append(
                "df.columns = [' '.join([str(c) for c in col]).strip() for col in df.columns.values]"
            )
        code.append('df = df.rename_axis(None, axis=1)')
        return '\n'.join(code)



import pandas as pd
import requests
from six import PY3, BytesIO, StringIO

from dtale.app import show
from dtale.cli.clickutils import get_loader_options, loader_prop_keys

'''
  IMPORTANT!!! These global variables are required for building any customized CLI loader.
  When build_loaders runs startup it will search for any modules containing the global variable LOADER_KEY.
'''
LOADER_KEY = 'csv'
LOADER_PROPS = [
    dict(name='path', help='path to CSV file'),
    dict(name='proxy', help="proxy URL if you're passing in a URL for --csv-path"),
    dict(name='parse_dates', help='comma-separated string of column names which should be parsed as dates')
]


# IMPORTANT!!! This function is required if you would like to be able to use this loader from the back-end.
def show_loader(**kwargs):
    return show(data_loader=lambda: loader_func(**kwargs), **kwargs)


def loader_func(**kwargs):
    path = kwargs.pop('path')
    if path.startswith('http://') or path.startswith('https://'):  # add support for URLs
        proxy = kwargs.pop('proxy', None)
        req_kwargs = {}
        if proxy is not None:
            req_kwargs['proxies'] = dict(http=proxy, https=proxy)
        resp = requests.get(path, **req_kwargs)
        assert resp.status_code == 200
        path = BytesIO(resp.content) if PY3 else StringIO(resp.content.decode('utf-8'))
    return pd.read_csv(path, **{k: v for k, v in kwargs.items() if k in loader_prop_keys(LOADER_PROPS)})


# IMPORTANT!!! This function is required for building any customized CLI loader.
def find_loader(kwargs):
    """
    CSV implementation of data loader which will return a function if any of the
    `click` options based on LOADER_KEY & LOADER_PROPS have been used, otherwise return None

    :param kwargs: Optional keyword arguments to be passed from `click`
    :return: data loader function for CSV implementation
    """
    csv_opts = get_loader_options(LOADER_KEY, kwargs)
    if len([f for f in csv_opts.values() if f]):
        def _csv_loader():
            csv_arg_parsers = {  # TODO: add additional arg parsers
                'parse_dates': lambda v: v.split(',') if v else None
            }
            kwargs = {k: csv_arg_parsers.get(k, lambda v: v)(v) for k, v in csv_opts.items()}
            return loader_func(**kwargs)
        return _csv_loader
    return None



from dash.development.base_component import Component, _explicitize_args


class Wordcloud(Component):
    """
    A Wordcloud component.
    Wordcloud is a wrapper component for react-wordcloud usage by dash.
    It takes a property, `data`, containing a series of words and a series of weights and displays a wordcloud
    or group of wordclouds depending on whether a `group` value has been specified.

    Keyword arguments:
    - id (string; required): The ID used to identify this component in Dash callbacks.
    - data (dict; optional): Server-side data containing words "data[group].x" and weights "data[group][y-prop]".
    - y (list of strings; optional): List of properties to use as weights.
    - group (list of strings; optional): List of properties to use as groups.
    - height (number; default 400): Height of wordcloud in pixels (default: 400).
    """
    @_explicitize_args
    def __init__(self, id=Component.REQUIRED, data=Component.UNDEFINED, y=Component.UNDEFINED,
                 group=Component.UNDEFINED, height=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'data', 'y', 'group', 'height']
        self._type = 'Wordcloud'
        self._namespace = 'components'
        self._valid_wildcard_attributes = []
        self.available_properties = ['id', 'data', 'y', 'group', 'height']
        self.available_wildcard_properties = []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in ['id']:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Wordcloud, self).__init__(**args)



from __future__ import division

import decimal
import json
import os
import socket
import sys
import time
import traceback
from builtins import map, object
from logging import getLogger

from flask import jsonify as _jsonify

import numpy as np
import pandas as pd
from past.utils import old_div
from six import PY3, StringIO

import dtale.global_state as global_state

logger = getLogger(__name__)


def running_with_pytest():
    """
    Checks to see if D-Tale has been initiated from test

    :return: `True` if executed from test, `False` otherwise
    :rtype: bool
    """
    return hasattr(sys, '_called_from_test')


def running_with_flask_debug():
    """
    Checks to see if D-Tale has been initiated from Flask

    :return: `True` if executed from test, `False` otherwise
    :rtype: bool
    """
    return os.environ.get('WERKZEUG_RUN_MAIN') == 'true'


def get_host(host=None):
    """
    Returns host input if it exists otherwise the output of :func:`python:socket.gethostname`

    :param host: hostname, can start with 'http://', 'https://' or just the hostname itself
    :type host: str, optional
    :return: str
    """

    def is_valid_host(host):
        try:
            socket.gethostbyname(host.split('://')[-1])
            return True
        except BaseException:
            return False

    if host is None:
        socket_host = socket.gethostname()
        if is_valid_host(socket_host):
            return socket_host
        return 'localhost'
    if is_valid_host(host):
        return host
    raise Exception('Hostname ({}) is not recognized'.format(host))


def build_url(port, host):
    """
    Returns full url combining host(if not specified will use the output of :func:`python:socket.gethostname`) & port

    :param port: integer string for the port to be used by the :class:`flask:flask.Flask` process
    :type port: str
    :param host: hostname, can start with 'http://', 'https://' or just the hostname itself
    :type host: str, optional
    :return: str
    """
    final_port = ':{}'.format(port) if port is not None else ''
    if (host or '').startswith('http'):
        return '{}{}'.format(host, final_port)
    return 'http://{}{}'.format(host, final_port)


def build_shutdown_url(base):
    """
    Builds the shutdown endpoint for the specified port

    :param port: integer string for a D-Tale process's port
    :type port: str
    :return: URL string of the shutdown endpoint for the current server and port passed
    """
    return '{}/shutdown'.format(base)


def get_str_arg(r, name, default=None):
    """
    Retrieve argument from :attr:`flask:flask.request` and convert to string

    :param r: :attr:`flask:flask.request`
    :param name: argument name
    :type: str
    :param default: default value if parameter is non-existent, defaults to None
    :return: string argument value
    """
    val = r.args.get(name)
    if val is None or val == '':
        return default
    else:
        try:
            return str(val)
        except BaseException:
            return default


def get_json_arg(r, name, default=None):
    """
    Retrieve argument from :attr:`flask:flask.request` and parse JSON to python data structure

    :param r: :attr:`flask:flask.request`
    :param name: argument name
    :type: str
    :param default: default value if parameter is non-existent, defaults to None
    :return: parsed JSON
    """
    val = r.args.get(name)
    if val is None or val == '':
        return default
    else:
        return json.loads(val)


def get_int_arg(r, name, default=None):
    """
    Retrieve argument from :attr:`flask:flask.request` and convert to integer

    :param r: :attr:`flask:flask.request`
    :param name: argument name
    :type: str
    :param default: default value if parameter is non-existent, defaults to None
    :return: integer argument value
    """
    val = r.args.get(name)
    if val is None or val == '':
        return default
    else:
        try:
            return int(val)
        except BaseException:
            return default


def get_float_arg(r, name, default=None):
    """
    Retrieve argument from :attr:`flask:flask.request` and convert to float

    :param r: :attr:`flask:flask.request`
    :param name: argument name
    :type: str
    :param default: default value if parameter is non-existent, defaults to None
    :return: float argument value
    """
    val = r.args.get(name)
    if val is None or val == '':
        return default
    else:
        try:
            return float(val)
        except BaseException:
            return default


def get_bool_arg(r, name):
    """
    Retrieve argument from :attr:`flask:flask.request` and convert to boolean

    :param r: :attr:`flask:flask.request`
    :param name: argument name
    :type: str
    :return: `True` if lowercase value equals 'true', `False` otherwise
    """
    return r.args.get(name, 'false').lower() == 'true'


def json_string(x, nan_display='', **kwargs):
    """
    convert value to string to be used within JSON output

    If a :class:`python.UnicodeEncodeError` occurs then :func:`python:str.encode` will be called on input

    :param x: value to be converted to string
    :param nan_display: if `x` is :attr:`numpy:numpy.nan` then return this value
    :return: string value
    :rtype: str
    """
    if x or x in [False, 0]:
        try:
            return str(x)
        except UnicodeEncodeError:
            return x.encode('utf-8')
        except BaseException as ex:
            logger.exception(ex)
    return nan_display


def json_int(x, nan_display='', as_string=False, fmt='{:,d}'):
    """
    Convert value to integer to be used within JSON output

    :param x: value to be converted to integer
    :param nan_display: if `x` is :attr:`numpy:numpy.nan` then return this value
    :param as_string: return integer as a formatted string (EX: 1,000,000)
    :return: integer value
    :rtype: int
    """
    try:
        if not np.isnan(x) and not np.isinf(x):
            return fmt.format(int(x)) if as_string else int(x)
        return nan_display
    except BaseException:
        return nan_display


# hack to solve issues with formatting floats with a precision more than 4 decimal points
# https://stackoverflow.com/questions/38847690/convert-float-to-string-without-scientific-notation-and-false-precision
DECIMAL_CTX = decimal.Context()
DECIMAL_CTX.prec = 20


def json_float(x, precision=2, nan_display='nan', inf_display='inf', as_string=False):
    """
    Convert value to float to be used within JSON output

    :param x: value to be converted to integer
    :param precision: precision of float to be returned
    :param nan_display: if `x` is :attr:`numpy:numpy.nan` then return this value
    :param inf_display: if `x` is :attr:`numpy:numpy.inf` then return this value
    :param as_string: return float as a formatted string (EX: 1,234.5643)
    :return: float value
    :rtype: float
    """
    try:
        if np.isinf(x):
            return inf_display
        if not np.isnan(x):
            output = float(round(x, precision))
            if as_string:
                str_output = format(DECIMAL_CTX.create_decimal(repr(x)), ',.{}f'.format(str(precision)))
                # drop trailing zeroes off & trailing decimal points if necessary
                return str_output.rstrip('0').rstrip('.')
            return output
        return nan_display
    except BaseException:
        return nan_display


def json_date(x, fmt='%Y-%m-%d %H:%M:%S', nan_display='', **kwargs):
    """
    Convert value to date string to be used within JSON output

    :param x: value to be converted to date string
    :param fmt: the data string formatting to be applied
    :param nan_display: if `x` is :attr:`numpy:numpy.nan` then return this value
    :return: date string value
    :rtype: str (YYYY-MM-DD)
    """
    try:
        # calling unique on a pandas datetime column returns numpy datetime64
        output = (pd.Timestamp(x) if isinstance(x, np.datetime64) else x).strftime(fmt)
        empty_time = ' 00:00:00'
        if output.endswith(empty_time):
            return output[:-1 * len(empty_time)]
        return output
    except BaseException:
        return nan_display


def json_timestamp(x, nan_display='', **kwargs):
    """
    Convert value to timestamp (milliseconds) to be used within JSON output

    :param x: value to be converted to milliseconds
    :param nan_display: if `x` is :attr:`numpy:numpy.nan` then return this value
    :return: millisecond value
    :rtype: bigint
    """
    try:
        output = (pd.Timestamp(x) if isinstance(x, np.datetime64) else x)
        output = int((time.mktime(output.timetuple()) + (old_div(output.microsecond, 1000000.0))) * 1000)
        return output
    except BaseException:
        return nan_display


class JSONFormatter(object):
    """
    Class for formatting dictionaries and lists of dictionaries into JSON compliant data

    :Example:

        >>> nan_display = 'nan'
        >>> f = JSONFormatter(nan_display)
        >>> f.add_int(1, 'a')
        >>> f.add_float(2, 'b')
        >>> f.add_string(3, 'c')
        >>> jsonify(f.format_dicts([dict(a=1, b=2.0, c='c')]))
    """

    def __init__(self, nan_display=''):
        self.fmts = []
        self.nan_display = nan_display

    def add_string(self, idx, name=None):
        self.fmts.append([idx, name, json_string])

    def add_int(self, idx, name=None, as_string=False):
        def f(x, nan_display):
            return json_int(x, nan_display=nan_display, as_string=as_string)
        self.fmts.append([idx, name, f])

    def add_float(self, idx, name=None, precision=6, as_string=False):
        def f(x, nan_display):
            return json_float(x, precision, nan_display=nan_display, as_string=as_string)
        self.fmts.append([idx, name, f])

    def add_timestamp(self, idx, name=None):
        self.fmts.append([idx, name, json_timestamp])

    def add_date(self, idx, name=None, fmt='%Y-%m-%d %H:%M:%S'):
        def f(x, nan_display):
            return json_date(x, fmt=fmt, nan_display=nan_display)
        self.fmts.append([idx, name, f])

    def add_json(self, idx, name=None):
        def f(x, nan_display):
            if x is None or pd.isnull(x):
                return None
            return x
        self.fmts.append([idx, name, f])

    def format_dict(self, lst):
        return {name: f(lst[idx], nan_display=self.nan_display) for idx, name, f in self.fmts}

    def format_dicts(self, lsts):
        return [self.format_dict(l) for l in lsts]

    def format_lists(self, df):
        return {
            name: [f(v, nan_display=self.nan_display) for v in df[name].values]
            for _idx, name, f in self.fmts
            if name in df.columns
        }

    def format_df(self, df):
        formatters = {col: f for _idx, col, f in self.fmts}
        cols = [col for col in df.columns if col in formatters]
        return pd.concat(
            [df[col].apply(lambda v: formatters[col](v, nan_display=self.nan_display)) for col in cols], axis=1
        )


def classify_type(type_name):
    """

    :param type_name: string label for value from :meth:`pandas:pandas.DataFrame.dtypes`
    :return: shortened string label for dtype
        S = str
        B = bool
        F = float
        I = int
        D = timestamp or datetime
        TD = timedelta
    :rtype: str
    """
    lower_type_name = (type_name or '').lower()
    if lower_type_name.startswith('str'):
        return 'S'
    if lower_type_name.startswith('bool'):
        return 'B'
    if lower_type_name.startswith('float'):
        return 'F'
    if lower_type_name.startswith('int'):
        return 'I'
    if any([t for t in ['timestamp', 'datetime'] if lower_type_name.startswith(t)]):
        return 'D'
    if lower_type_name.startswith('timedelta'):
        return 'TD'
    return 'S'


def retrieve_grid_params(req, props=None):
    """
    Pull out grid parameters from :attr:`flask:flask.request` arguments and return as a `dict`

    :param req: :attr:`flask:flask.request`
    :param props: argument names
    :type props: list
    :return: dictionary of argument/value pairs
    :rtype: dict
    """
    params = dict()
    params['sort_column'] = get_str_arg(req, 'sortColumn')
    params['sort_direction'] = get_str_arg(req, 'sortDirection')
    sort = get_str_arg(req, 'sort')
    if sort:
        params['sort'] = json.loads(sort)
    return params


def sort_df_for_grid(df, params):
    """
    Sort dataframe based on 'sort' property in parameter dictionary. Sort
    configuration is of the following shape:
    {
        sort: [
            [col1, ASC],
            [col2, DESC],
            ...
        ]
    }

    :param df: dataframe
    :type df: :class:`pandas:pandas.DataFrame`
    :param params: arguments from :attr:`flask:flask.request`
    :type params: dict
    :return: sorted dataframe
    :rtype: :class:`pandas:pandas.DataFrame`
    """
    if 'sort' in params:
        cols, dirs = [], []
        for col, dir in params['sort']:
            cols.append(col)
            dirs.append(dir == 'ASC')
        return df.sort_values(cols, ascending=dirs)
    return df.sort_index()


def find_dtype(s):
    """
    Helper function to determine the dtype of a :class:`pandas:pandas.Series`
    """
    if s.dtype.name == 'object':
        return pd.api.types.infer_dtype(s, skipna=True)
    else:
        return s.dtype.name


def get_dtypes(df):
    """
    Build dictionary of column/dtype name pairs from :class:`pandas:pandas.DataFrame`
    """
    def _load():
        for col in df.columns:
            yield col, find_dtype(df[col])

    return dict(list(_load()))


def grid_columns(df):
    """
    Build list of {name, dtype} dictionaries for columns in :class:`pandas:pandas.DataFrame`
    """
    data_type_info = get_dtypes(df)
    return [dict(name=c, dtype=data_type_info[c]) for c in df.columns]


DF_MAPPINGS = {
    'I': lambda f, i, c: f.add_int(i, c),
    'D': lambda f, i, c: f.add_date(i, c),
    'F': lambda f, i, c: f.add_float(i, c),
    'S': lambda f, i, c: f.add_string(i, c)
}


def find_dtype_formatter(dtype, overrides=None):
    type_classification = classify_type(dtype)
    if type_classification in (overrides or {}):
        return overrides[type_classification]
    if type_classification == 'I':
        return json_int
    if type_classification == 'D':
        return json_date
    if type_classification == 'F':
        return json_float
    return json_string


def grid_formatter(col_types, nan_display='', overrides=None):
    """
    Build :class:`dtale.utils.JSONFormatter` from :class:`pandas:pandas.DataFrame`
    """
    f = JSONFormatter(nan_display)
    mappings = dict_merge(DF_MAPPINGS, overrides or {})
    for i, ct in enumerate(col_types, 1):
        c, dtype = map(ct.get, ['name', 'dtype'])
        type_classification = classify_type(dtype)
        mappings.get(type_classification, DF_MAPPINGS['S'])(f, i, c)
    return f


def format_grid(df):
    """
    Translate :class:`pandas:pandas.DataFrame` to well-formed JSON.  Structure is as follows:
    {
        results: [
            {col1: val1_row1,...,colN: valN_row1},
            ...,
            {col1: val1_rowN,...,colN: valN_rowN},
        ],
        columns: [
            {name: col1, dtype: int},
            ...,
            {name: colN, dtype: float},
        ]
    }

    :param df: dataframe
    :type df: :class:`pandas:pandas.DataFrame`
    :return: JSON
    """
    col_types = grid_columns(df)
    f = grid_formatter(col_types)
    return {
        'results': f.format_dicts(df.itertuples()),
        'columns': col_types
    }


def handle_error(error_info):
    """
    Boilerplate exception messaging
    """
    logger.exception("Exception occurred while processing request: {}".format(error_info.get('error')))


def jsonify(return_data={}, **kwargs):
    """
    Overriding Flask's jsonify method to account for extra error handling

    :param return_data: dictionary of data to be passed to :meth:`flask:flask.jsonify`
    :param kwargs: Optional keyword arguments merged into return_data
    :return: output of :meth:`flask:flask.jsonify`
    """
    if isinstance(return_data, dict) and return_data.get('error'):
        handle_error(return_data)
        return _jsonify(dict_merge(dict(success=False), dict_merge(kwargs, return_data)))
    if len(kwargs):
        return _jsonify(dict_merge(kwargs, return_data))
    return _jsonify(return_data)


class ChartBuildingError(Exception):
    """
    Exception for signalling there was an issue constructing the data for your chart.
    """
    def __init__(self, error, details=None):
        super(ChartBuildingError, self).__init__('Chart Error')
        self.error = error
        self.details = details


def jsonify_error(e):
    tb = traceback.format_exc()
    if isinstance(e, ChartBuildingError):
        if e.details:
            tb = e.details
        e = e.error
    return jsonify(dict(error=str(e), traceback=str(tb)))


def find_selected_column(data, col):
    """
    In case we come across a series which after reset_index()
    it's columns are [date, security_id, values]
    in which case we want the last column

    :param data: dataframe
    :type data: :class:`pandas:pandas.DataFrame`
    :param col: column name
    :type col: str
    :return: column name if it exists within the dataframe's columns, the last column within the dataframe otherwise
    :rtype: str
    """

    return col if col in data.columns else data.columns[-1]


def make_list(vals):
    """
    Convert a value that is optionally list or scalar
    into a list
    """
    if vals is None:
        return []
    elif isinstance(vals, (list, tuple)):
        return vals
    return [vals]


def dict_merge(d1, d2, *args):
    """
    Merges two dictionaries.  Items of the second dictionary will
    replace items of the first dictionary if there are any overlaps.
    Either dictionary can be None.  An empty dictionary {} will be
    returned if both dictionaries are None.

    :param d1: First dictionary can be None
    :type d1: dict
    :param d2: Second dictionary can be None
    :type d1: dict
    :return: new dictionary with the contents of d2 overlaying the contents of d1
    :rtype: dict
    """
    def _dict_merge(d11, d12):
        if not d11:
            return d12 or {}
        elif not d12:
            return d11 or {}
        return dict(list(d11.items()) + list(d12.items()))
    ret = _dict_merge(d1, d2)
    for d in args:
        ret = _dict_merge(ret, d)
    return ret


def flatten_lists(lists):
    """
    Take an iterable containing iterables and flatten them into one list.
        - [[1], [2], [3, 4]] => [1, 2, 3, 4]
    """
    return [item for sublist in lists for item in sublist]


def divide_chunks(l, n):
    """
    Break list input 'l' up into smaller lists of size 'n'
    """
    # looping till length l
    for i in range(0, len(l), n):
        yield l[i:i + n]


def build_query(data_id, query=None):
    curr_settings = global_state.get_settings(data_id) or {}
    return inner_build_query(curr_settings, query)


def inner_build_query(settings, query=None):
    query_segs = []
    for p in ['columnFilters', 'outlierFilters']:
        curr_filters = settings.get(p) or {}
        for col, filter_cfg in curr_filters.items():
            query_segs.append(filter_cfg['query'])
    if query not in [None, '']:
        query_segs.append(query)
    return ' and '.join(query_segs)


def run_query(df, query, context_vars=None, ignore_empty=False):
    """
    Utility function for running :func:`pandas:pandas.DataFrame.query` . This function contains extra logic to
    handle when column names contain special characters.  Looks like pandas will be handling this in a future
    version: https://github.com/pandas-dev/pandas/issues/27017

    The logic to handle these special characters in the meantime is only available in Python 3+

    :param df: input dataframe
    :type df: :class:`pandas:pandas.DataFrame`
    :param query: query string
    :type query: str
    :param context_vars: dictionary of user-defined variables which can be referenced by name in query strings
    :type context_vars: dict, optional
    :return: filtered dataframe
    """
    if (query or '') == '':
        return df

    # https://stackoverflow.com/a/40083013/12616360 only supporting filter cleanup for python 3+
    if PY3:
        invalid_column_names = [x for x in df.columns.values if not x.isidentifier()]

        # Make replacements in the query and keep track
        # NOTE: This method fails if the frame has columns called REPL_0 etc.
        replacements = dict()
        final_query = str(query)
        for cn in invalid_column_names:
            r = 'REPL_{}'.format(str(invalid_column_names.index(cn)))
            final_query = final_query.replace(cn, r)
            replacements[cn] = r

        inv_replacements = {replacements[k]: k for k in replacements.keys()}
        df = df.rename(columns=replacements)

        df = df.query(final_query, local_dict=context_vars or {})
        df = df.rename(columns=inv_replacements)
    else:
        df = df.query(query, local_dict=context_vars or {})

    if not len(df) and not ignore_empty:
        raise Exception('query "{}" found no data, please alter'.format(query))
    return df


class DuplicateDataError(Exception):
    """
    Exception for signalling that similar data is trying to be loaded to D-Tale again.  Is this correct?
    """
    def __init__(self, data_id):
        super(DuplicateDataError, self).__init__("Duplicate Data")
        self.data_id = data_id


def build_code_export(data_id, imports='import pandas as pd\n\n', query=None):
    """
    Helper function for building a string representing the code that was run to get the data you are viewing to that
    point.

    :param data_id: integer string identifier for a D-Tale process's data
    :type data_id: str
    :param imports: string representing the imports at the top of the code string
    :type imports: string, optional
    :param query: pandas dataframe query string
    :type query: str, optional
    :return: python code string
    """
    history = global_state.get_history(data_id) or []
    settings = global_state.get_settings(data_id) or {}
    ctxt_vars = global_state.get_context_variables(data_id)

    startup_code = settings.get('startup_code')
    startup_code = '# Data Re-shaping\n{}\n\n'.format(startup_code) if startup_code else ''
    startup_str = (
        "# DISCLAIMER: 'df' refers to the data you passed in when calling 'dtale.show'\n\n"
        '{imports}'
        '{startup}'
        'if isinstance(df, (pd.DatetimeIndex, pd.MultiIndex)):\n'
        '\tdf = df.to_frame(index=False)\n\n'
        '# remove any pre-existing indices for ease of use in the D-Tale code, but this is not required\n'
        "df = df.reset_index().drop('index', axis=1, errors='ignore')\n"
        'df.columns = [str(c) for c in df.columns]  # update columns to strings in case they are numbers\n'
    ).format(imports=imports, startup=startup_code)
    final_history = [startup_str] + history
    final_query = query
    if final_query is None:
        final_query = settings.get('query')

    if final_query is not None:
        if len(ctxt_vars or {}):
            final_history.append((
                "\n# this is injecting any context variables you may have passed into 'dtale.show'\n"
                "import dtale.global_state as dtale_global_state\n"
                "\n# DISCLAIMER: running this line in a different process than the one it originated will produce\n"
                "#             differing results\n"
                "ctxt_vars = dtale_global_state.get_context_variables('{data_id}')\n\n"
                'df = df.query("{query}", local_dict=ctxt_vars)\n'
            ).format(query=final_query, data_id=data_id))
        else:
            final_history.append("df = df.query('{}')\n".format(final_query))
    elif 'query' in settings:
        final_history.append("df = df.query('{}')\n".format(settings['query']))
    if 'sort' in settings:
        cols, dirs = [], []
        for col, dir in settings['sort']:
            cols.append(col)
            dirs.append('True' if dir == 'ASC' else 'False')
        final_history.append("df = df.sort_values(['{cols}'], ascending=[{dirs}])\n".format(
            cols=', '.join(cols), dirs="', '".join(dirs)
        ))
    return final_history


def export_to_csv_buffer(data, tsv=False):
    kwargs = dict(encoding='utf-8', index=False)
    if tsv:
        kwargs['sep'] = '\t'
    csv_buffer = StringIO()
    data.to_csv(csv_buffer, **kwargs)
    csv_buffer.seek(0)
    return csv_buffer



from logging import getLogger

import pandas as pd

from dtale.app import show
from dtale.cli.clickutils import get_loader_options

logger = getLogger(__name__)

'''
  IMPORTANT!!! This global variable is required for building any customized CLI loader.
  When find loaders on startup it will search for any modules containing the global variable LOADER_KEY.
'''
NOW = pd.Timestamp('now').strftime('%Y%m%d')
LOADER_KEY = 'arctic'
LOADER_PROPS = [
    dict(name='host', help='arctic hostname'),
    dict(name='library', help='library within --arctic-host'),
    dict(name='node', help='node within --arctic-library'),
    dict(name='start', help='start-date of range to load if reading from ChunkStore (EX: {})'.format(NOW)),
    dict(name='end', help='end-date of range to load if reading from ChunkStore (EX: {})'.format(NOW)),
]


# IMPORTANT!!! This function is required if you would like to be able to use this loader from the back-end.
def show_loader(**kwargs):
    return show(data_loader=lambda: loader_func(**kwargs), **kwargs)


def loader_func(**kwargs):
    try:
        from arctic import Arctic
        from arctic.store.versioned_item import VersionedItem
    except ImportError:
        raise ImportError('In order to use the arctic loader you must install arctic!')
    host = Arctic(kwargs.get('host'))
    lib = host.get_library(kwargs.get('library'))
    read_kwargs = {}
    start, end = (kwargs.get(p) for p in ['start', 'end'])
    if start and end:
        read_kwargs['chunk_range'] = pd.date_range(start, end)
    data = lib.read(kwargs.get('node'), **read_kwargs)
    if isinstance(data, VersionedItem):
        data = data.data
    return data


# IMPORTANT!!! This function is required for building any customized CLI loader.
def find_loader(kwargs):
    """
    Arctic implementation of data loader which will return a function if any of the
    `click` options based on LOADER_KEY & LOADER_PROPS have been used, otherwise return None

    :param kwargs: Optional keyword arguments to be passed from `click`
    :return: data loader function for arctic implementation
    """
    arctic_opts = get_loader_options(LOADER_KEY, kwargs)
    if len([f for f in arctic_opts.values() if f]):

        def _arctic_loader():
            return loader_func(**arctic_opts)

        return _arctic_loader
    return None



from logging import getLogger

import click

from dtale.app import find_free_port, show
from dtale.cli.clickutils import (LOG_LEVELS, get_log_options, run,
                                  setup_logging)
from dtale.cli.loaders import check_loaders, setup_loader_options

logger = getLogger(__name__)


@click.command(name='main', help='Run D-Tale from command-line')
@click.option('--host', type=str, help='hostname or IP address of process')
@click.option('--port', type=int, help='port number of process')
@click.option('--debug', is_flag=True, help="flag to switch on Flask's debug mode")
@click.option('--no-reaper', is_flag=True,
              help='flag to turn off auto-reaping (process cleanup after period of inactivity)')
@click.option('--open-browser', is_flag=True, help='flag to automatically open default web browser on startup')
@click.option('--name', type=str, help='name to apply to your D-Tale session')
@setup_loader_options()
@click.option('--log', 'logfile', help='Log file name')
@click.option('--log-level',
              help='Set the logging level',
              type=click.Choice(list(LOG_LEVELS.keys())),
              default='info',
              show_default=True)
@click.option('-v', '--verbose', help='Set the logging level to debug', is_flag=True)
def main(host=None, port=None, debug=False, no_reaper=False, open_browser=False, name=None, **kwargs):
    """
    Runs a local server for the D-Tale application.

    This local server is recommended when you have a pandas object stored in a CSV
    or retrievable from :class:`arctic.Arctic` data store.
    """
    log_opts = get_log_options(kwargs)
    setup_logging(log_opts.get('logfile'), log_opts.get('log_level'), log_opts.get('verbose'))

    data_loader = check_loaders(kwargs)

    show(host=host, port=int(port or find_free_port()), debug=debug, subprocess=False, data_loader=data_loader,
         reaper_on=not no_reaper, open_browser=open_browser, name=name, **kwargs)


if __name__ == '__main__':
    run(main)



import logging
import socket
import sys
from builtins import str
from datetime import datetime

import click
import pkg_resources
from six import string_types

logger = logging.getLogger(__name__)

LOG_LEVELS = dict(
    debug=logging.DEBUG, info=logging.INFO, warning=logging.WARNING, error=logging.ERROR, critical=logging.CRITICAL
)


def setup_logging(logfile, log_level, verbose=False):
    """
    Utility method for setting up logging configuration

    :param logfile: location of D-Tale logs
    :type logfile: str
    :param log_level: D-Tale's implementation of standard logging levels
    :type log_level: str, options are debug, info, warning, error or, critical
    :param verbose: turns on verbose logging, defaults to False
    :type verbose: bool, optional
    :return:
    """

    if log_level == 'verbose' or verbose:
        log_level = LOG_LEVELS['debug']
    elif log_level:
        log_level = LOG_LEVELS[log_level]
    else:
        log_level = LOG_LEVELS['info']

    logging.getLogger().handlers = []

    fmt = "%(asctime)s - %(levelname)-8s - %(message)s"
    logging.basicConfig(format=fmt, level=log_level)

    if logfile:
        fh = logging.FileHandler(logfile, mode='w')
        fh.setFormatter(logging.Formatter(fmt))
        logging.getLogger().addHandler(fh)

    for handler in logging.getLogger().handlers:
        handler.setLevel(log_level)

    logger.debug("{}".format(' '.join(sys.argv)))
    try:
        logger.debug("Hostname: {}".format(socket.gethostname()))
    except Exception as e:
        logger.exception(e)


def loader_options(key, params):
    """
    Builds a decorator which will append a list of click parameters based on
    a combination of prefix & parameter name

    loader_options('foo', ['bar', 'baz']) => click.option('--foo-bar'), click.option('--foo-baz')

    :param key: parameter prefix
    :type key: str
    :param params: parameters names
    :type params: list
    :return: filter
    :rtype: func
    """
    def decorator(f):
        for p in params:
            f = click.option('--' + key + '-' + p, help='Override {} {}'.format(key, p))(f)
        return f
    return decorator


def get_loader_options(key, options):
    """
    Filters dictionary of click parameters for ones which start with a certain prefix

    :param key: click option prefix
    :type key: str
    :param options: click options
    :type options: dict
    :return: dictionary of click options with start with key
    :rtype: dict
    """

    def _build_key(option):
        segs = option.split('_')
        if len(segs) == 1:
            return ''
        return '_'.join(segs[1:])
    return dict(((_build_key(k), v) for k, v in options.items() if k.startswith(key) if v is not None))


def get_log_options(options):
    """
    Click logging options (logfile, log_level, verbose)
    """
    names = ['logfile', 'log_level', 'verbose']
    return get_named_options(names, options)


def get_named_options(names, options):
    """
    Filters click options for options in a specific subset of names
    """
    return {k: v for k, v in options.items() if k in names}


def retrieve_meta_info_and_version(name):
    """
    Retrieves meta information and versioning of specific package name
    """
    try:
        dist = pkg_resources.get_distribution(name)
    except BaseException:
        return None, 'unknown'
    try:
        for line in dist._get_metadata(dist.PKG_INFO):
            if line.startswith('Description:'):
                return line[len('Description:'):].strip(), dist.version
    except BaseException:
        pass
    return None, dist.version


def get_args(click_wrapper):
    """
    Retrieves arguments being passed in from the command-line and applying them to a click command

    :param click_wrapper: click context
    :return: list of arguments
    :rtype: list
    """

    # Get the options defined on cli
    params = []
    for opt in click_wrapper.params:
        params.extend(opt.opts)

    args = sys.argv[1:]

    # Reorder the args so that cli options come before
    # subcommand options. This is a click limitation.
    head, tail = [], []
    for a in args:
        if a in params or head and head[-1] in params:
            head.append(a)
        else:
            tail.append(a)
    return head + tail


def run(click_wrapper):
    """
    Wrapper function for executing click commands and handling exceptions produced from
    click commands with versioning boilerplate

    :param click_wrapper: click context
    """

    t1 = datetime.now()
    try:
        try:
            dtale_name = 'dtale'
            bld_info, ver_info = retrieve_meta_info_and_version(dtale_name)
            logger.debug('{} bld: {}'.format(dtale_name, bld_info))
            logger.debug('{} ver: {}'.format(dtale_name, ver_info))
        except BaseException:
            logger.debug('failure to retrieve metadata for: {}'.format(dtale_name))

        args = get_args(click_wrapper)
        click_wrapper(args)
        sys.exit(0)
    except Exception as ex:
        logger.exception("Fatal error: " + str(ex))
        sys.exit(1)
    finally:
        logger.info("Elapsed time: %s" % (datetime.now() - t1))


def loader_prop_keys(prop_cfgs):
    return [p if isinstance(p, string_types) else p['name'] for p in prop_cfgs]



from six import PY3
from six.moves.collections_abc import MutableMapping

DATA = {}
DTYPES = {}
SETTINGS = {}
METADATA = {}
CONTEXT_VARIABLES = {}
HISTORY = {}


def get_data(data_id=None):
    global DATA

    if data_id is None:
        return _as_dict(DATA)
    return DATA.get(data_id)


def get_dtypes(data_id=None):
    global DTYPES

    if data_id is None:
        return _as_dict(DTYPES)
    return DTYPES.get(data_id)


def get_settings(data_id=None):
    global SETTINGS

    if data_id is None:
        return _as_dict(SETTINGS)
    return SETTINGS.get(data_id)


def get_metadata(data_id=None):
    global METADATA

    if data_id is None:
        return _as_dict(METADATA)
    return METADATA.get(data_id)


def get_context_variables(data_id=None):
    global CONTEXT_VARIABLES

    if data_id is None:
        return _as_dict(CONTEXT_VARIABLES)
    return CONTEXT_VARIABLES.get(data_id)


def get_history(data_id=None):
    global HISTORY

    if data_id is None:
        return _as_dict(HISTORY)
    return HISTORY.get(data_id)


def set_data(data_id, val):
    global DATA

    DATA[data_id] = val


def set_dtypes(data_id, val):
    global DTYPES

    DTYPES[data_id] = val


def set_settings(data_id, val):
    global SETTINGS

    SETTINGS[data_id] = val


def set_metadata(data_id, val):
    global METADATA

    METADATA[data_id] = val


def set_context_variables(data_id, val):
    global CONTEXT_VARIABLES

    CONTEXT_VARIABLES[data_id] = val


def set_history(data_id, val):
    global HISTORY

    HISTORY[data_id] = val


def cleanup(data_id=None):
    """
    Helper function for cleanup up state related to a D-Tale process with a specific port

    :param port: integer string for a D-Tale process's port
    :type port: str
    """
    global DATA, DTYPES, SETTINGS, METADATA, CONTEXT_VARIABLES, HISTORY

    if data_id is None:
        DATA.clear()
        SETTINGS.clear()
        DTYPES.clear()
        METADATA.clear()
        CONTEXT_VARIABLES.clear()
        HISTORY.clear()
    else:
        for store in [DATA, DTYPES, SETTINGS, METADATA, CONTEXT_VARIABLES, HISTORY]:
            if data_id in store:
                del store[data_id]


def _as_dict(store):
    """Return the dict representation of a data store.
    Stores must either be an instance of MutableMapping OR have a to_dict method.

    :param store: data store (dict, redis connection, etc.)
    :return: dict
    """
    return dict(store) if isinstance(store, MutableMapping) else store.to_dict()


def use_store(store_class, create_store):
    """
    Customize how dtale stores and retrieves global data.
    By default it uses global dictionaries, but this can be problematic if
    there are memory limitations or multiple python processes are running.
    Ex: a web server with multiple workers (processes) for processing requests.

    :param store_class: Class providing an interface to the data store. To be valid, it must:
                        1. Implement get, clear, __setitem__, __delitem__, __iter__, __len__, __contains__.
                        2. Either be a subclass of MutableMapping or implement the 'to_dict' method.
    :param create_store: Factory function for producing instances of <store_class>.
                         Must take 'name' as the only parameter.
    :return: None
    """
    import inspect

    assert inspect.isclass(store_class), 'Must be a class'
    assert all(hasattr(store_class, a) for a in (
        'get', 'clear', '__setitem__', '__delitem__', '__len__', '__contains__')), 'Missing required methods'
    assert (issubclass(store_class, MutableMapping)
            or hasattr(store_class, 'to_dict')), 'Must subclass MutableMapping or implement "to_dict"'

    assert inspect.isfunction(create_store), 'Must be a function'
    if PY3:
        assert list(inspect.signature(create_store).parameters) == ['name'], 'Must take "name" as the only parameter'
    else:
        assert inspect.getargspec(create_store).args == ['name'], 'Must take "name" as the only parameter'

    def convert(old_store, name):
        """Convert a data store to the new type

        :param old_store: old data store
        :param name: name associated with this data store
        :return: new data store
        """
        new_store = create_store(name)
        assert isinstance(new_store, store_class)
        new_store.clear()
        for k, v in _as_dict(old_store).items():
            new_store[k] = v
        old_store.clear()
        return new_store

    global DATA, DTYPES, SETTINGS, METADATA, CONTEXT_VARIABLES, HISTORY

    DATA = convert(DATA, 'DATA')
    DTYPES = convert(DTYPES, 'DTYPES')
    SETTINGS = convert(SETTINGS, 'SETTINGS')
    METADATA = convert(METADATA, 'METADATA')
    CONTEXT_VARIABLES = convert(CONTEXT_VARIABLES, 'CONTEXT_VARIABLES')
    HISTORY = convert(HISTORY, 'HISTORY')


def use_default_store():
    """Use the default global data store, which is dictionaries in memory."""
    def create_dict(name):
        return dict()

    use_store(dict, create_dict)


def use_shelve_store(directory):
    """
    Configure dtale to use python's standard 'shelve' library for a persistent global data store.

    :param directory: directory that the shelve db files will be stored in
    :type directory: str
    :return: None
    """
    import shelve
    from os.path import join
    from functools import wraps

    def with_db_access(flag):
        """
        Factory for creating decorators that will handle opening the database file
        with the given access and then close it after executing the method.
        """
        def decorator(func):
            @wraps(func)
            def wrapper(self, *args, **kwargs):
                self.db = shelve.open(self.filename, flag=flag)
                result = func(self, *args, **kwargs)
                self.db.close()
                return result
            return wrapper
        return decorator

    read = with_db_access('r')
    write = with_db_access('w')

    class DtaleShelf:
        """Interface allowing dtale to use 'shelf' databases for global data storage."""
        def __init__(self, filename):
            self.filename = filename
            self.db = shelve.open(self.filename, flag='n')
            self.db.close()

        @read
        def get(self, key):
            return self.db.get(key)

        @write
        def __setitem__(self, key, value):
            self.db[key] = value

        @write
        def __delitem__(self, key):
            del self.db[key]

        @read
        def __contains__(self, key):
            return key in self.db

        @write
        def clear(self):
            self.db.clear()

        @read
        def to_dict(self):
            return dict(self.db)

        @read
        def __len__(self):
            return len(self.db)

    def create_shelf(name):
        file_path = join(directory, name)
        return DtaleShelf(file_path)

    use_store(DtaleShelf, create_shelf)


def use_redis_store(directory, *args, **kwargs):
    """Configure dtale to use redis for the global data store. Useful for web servers.

    :param db_folder: folder that db files will be stored in
    :type db_folder: str
    :param args: All other arguments supported by the redislite.Redis() class
    :param kwargs: All other keyword arguments supported by the redislite.Redis() class
    :return: None
    """
    import pickle
    from os.path import join
    try:
        from redislite import Redis
    except ImportError:
        raise Exception('redislite must be installed')

    class DtaleRedis(Redis):
        """Wrapper class around Redis() to make it work as a global data store in dtale."""
        def get(self, name, *args, **kwargs):
            value = super(DtaleRedis, self).get(name, *args, **kwargs)
            if value is not None:
                return pickle.loads(value)

        def set(self, name, value, *args, **kwargs):
            value = pickle.dumps(value)
            return super(DtaleRedis, self).set(name, value, *args, **kwargs)

        def clear(self):
            self.flushdb()

        def to_dict(self):
            return {k.decode('utf-8'): self.get(k) for k in self.keys()}

        def __len__(self):
            return len(self.keys())

    def create_redis(name):
        file_path = join(directory, name + '.db')
        return DtaleRedis(file_path, *args, **kwargs)

    use_store(DtaleRedis, create_redis)



import random
import string
import time

import numpy as np
import pandas as pd

import dtale.global_state as global_state
from dtale.utils import classify_type


class ColumnBuilder(object):

    def __init__(self, data_id, column_type, name, cfg):
        self.data_id = data_id
        if column_type == 'numeric':
            self.builder = NumericColumnBuilder(name, cfg)
        elif column_type == 'datetime':
            self.builder = DatetimeColumnBuilder(name, cfg)
        elif column_type == 'bins':
            self.builder = BinsColumnBuilder(name, cfg)
        elif column_type == 'random':
            self.builder = RandomColumnBuilder(name, cfg)
        elif column_type == 'type_conversion':
            self.builder = TypeConversionColumnBuilder(name, cfg)
        else:
            raise NotImplementedError("'{}' column builder not implemented yet!".format(column_type))

    def build_column(self):
        data = global_state.get_data(self.data_id)
        return self.builder.build_column(data)

    def build_code(self):
        return self.builder.build_code()


class NumericColumnBuilder(object):

    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        left, right, operation = (self.cfg.get(p) for p in ['left', 'right', 'operation'])
        left = data[left['col']] if 'col' in left else float(left['val'])
        right = data[right['col']] if 'col' in right else float(right['val'])
        if operation == 'sum':
            return left + right
        if operation == 'difference':
            return left - right
        if operation == 'multiply':
            return left * right
        if operation == 'divide':
            return left / right
        return np.nan

    def build_code(self):
        left, right, operation = (self.cfg.get(p) for p in ['left', 'right', 'operation'])
        operations = dict(sum='+', difference='-', multiply='*', divide='/')
        return "df.loc[:, '{name}'] = {left} {operation} {right}".format(
            name=self.name,
            operation=operations.get(operation),
            left="df['{}']".format(left['col']) if 'col' in left else left['val'],
            right="df['{}']".format(right['col']) if 'col' in right else right['val']
        )


FREQ_MAPPING = dict(month='M', quarter='Q', year='Y')


class DatetimeColumnBuilder(object):

    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col = self.cfg['col']
        if 'property' in self.cfg:
            return getattr(data[col].dt, self.cfg['property'])
        conversion_key = self.cfg['conversion']
        [freq, how] = conversion_key.split('_')
        freq = FREQ_MAPPING[freq]
        conversion_data = data[[col]].set_index(col).index.to_period(freq).to_timestamp(how=how).normalize()
        return pd.Series(conversion_data, index=data.index, name=self.name)

    def build_code(self):
        if 'property' in self.cfg:
            return "df.loc[:, '{name}'] = df['{col}'].dt.{property}".format(name=self.name, **self.cfg)
        conversion_key = self.cfg['conversion']
        [freq, how] = conversion_key.split('_')
        freq = FREQ_MAPPING[freq]
        return (
            "{name}_data = data[['{col}']].set_index('{col}').index.to_period('{freq}')'"
            ".to_timestamp(how='{how}').normalize()\n"
            "df.loc[:, '{name}'] = pd.Series({name}_data, index=df.index, name='{name}')"
        ).format(name=self.name, col=self.cfg['col'], freq=freq, how=how)


class BinsColumnBuilder(object):

    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, operation, bins, labels = (self.cfg.get(p) for p in ['col', 'operation', 'bins', 'labels'])
        bins = int(bins)
        if operation == 'cut':
            bin_data = pd.cut(data[col], bins=bins)
        else:
            bin_data = pd.qcut(data[col], q=bins)
        if labels:
            cats = {idx: str(cat) for idx, cat in enumerate(labels.split(','))}
        else:
            cats = {idx: str(cat) for idx, cat in enumerate(bin_data.cat.categories)}
        return pd.Series(bin_data.cat.codes.map(cats), index=data.index, name=self.name)

    def build_code(self):
        col, operation, bins, labels = (self.cfg.get(p) for p in ['col', 'operation', 'bins', 'labels'])
        bins_code = []
        if operation == 'cut':
            bins_code.append("{name}_data = pd.cut(df['{col}'], bins={bins})".format(
                name=self.name, col=col, bins=bins
            ))
        else:
            bins_code.append("{name}_data = pd.qcut(df['{col}'], bins={bins})".format(
                name=self.name, col=col, bins=bins
            ))
        if labels:
            labels_str = ', '.join(['{}: {}'.format(idx, cat) for idx, cat in enumerate(labels.split(','))])
            labels_str = '{' + labels_str + '}'
            bins_code.append('{name}_cats = {labels}'.format(name=self.name, labels=labels_str))
        else:
            bins_code.append(
                '{name}_cats = {idx: str(cat) for idx, cat in enumerate({name}_data.cat.categories)}'
            )
        s_str = "df.loc[:, '{name}'] = pd.Series({name}_data.cat.codes.map({name}_cats), index=df.index, name='{name}')"
        bins_code.append(s_str.format(name=self.name))
        return '\n'.join(bins_code)


def id_generator(size=10, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(int(size)))


class RandomColumnBuilder(object):

    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        rand_type = self.cfg['type']
        if 'string' == rand_type:
            kwargs = dict(size=self.cfg.get('length', 10))
            if self.cfg.get('chars'):
                kwargs['chars'] = self.cfg['chars']
            return pd.Series(
                [id_generator(**kwargs) for _ in range(len(data))], index=data.index, name=self.name
            )
        if 'int' == rand_type:
            low = self.cfg.get('low', 0)
            high = self.cfg.get('high', 100)
            return pd.Series(
                np.random.randint(low, high=high, size=len(data)), index=data.index, name=self.name
            )
        if 'date' == rand_type:
            start = pd.Timestamp(self.cfg.get('start') or '19000101')
            end = pd.Timestamp(self.cfg.get('end') or '21991231')
            business_days = self.cfg.get('businessDay') is True
            timestamps = self.cfg.get('timestamps') is True
            if timestamps:
                def pp(start, end, n):
                    start_u = start.value // 10 ** 9
                    end_u = end.value // 10 ** 9
                    return pd.DatetimeIndex(
                        (10 ** 9 * np.random.randint(start_u, end_u, n)).view('M8[ns]')
                    )

                dates = pp(pd.Timestamp(start), pd.Timestamp(end), len(data))
            else:
                dates = pd.date_range(start, end, freq='B' if business_days else 'D')
                dates = [dates[i] for i in np.random.randint(0, len(dates) - 1, size=len(data))]
            return pd.Series(dates, index=data.index, name=self.name)
        if 'bool' == rand_type:
            return pd.Series(np.random.choice([True, False], len(data)), index=data.index, name=self.name)
        if 'choice' == rand_type:
            choices = self.cfg.get('choices') or 'a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z'
            choices = choices.split(',')
            return pd.Series(np.random.choice(choices, len(data)), index=data.index, name=self.name)

        # floats
        low = self.cfg.get('low', 0)
        high = self.cfg.get('high', 1)
        return pd.Series(np.random.uniform(low, high, len(data)), index=data.index, name=self.name)

    def build_code(self):
        rand_type = self.cfg['type']
        if 'string' == rand_type:
            kwargs = []
            if self.cfg.get('length') != 10:
                kwargs.append('size={size}'.format(size=self.cfg.get('length')))
            if self.cfg.get('chars'):
                kwargs.append("chars='{chars}'".format(chars=self.cfg.get('chars')))
            kwargs = ', '.join(kwargs)
            return (
                'import number\nimport random\n\n'
                'def id_generator(size=1500, chars=string.ascii_uppercase + string.digits):\n'
                "\treturn ''.join(random.choice(chars) for _ in range(size))\n\n"
                "df.loc[:, '{name}'] = pd.Series([id_generator({kwargs}) for _ in range(len(df)], index=df.index)"
            ).format(kwargs=kwargs, name=self.name)
            return "df.loc[:, '{name}'] = df['{col}'].dt.{property}".format(name=self.name, **self.cfg)

        if 'bool' == rand_type:
            return (
                "df.loc[:, '{name}'] = pd.Series(np.random.choice([True, False], len(df)), index=data.index"
            ).format(name=self.name)
        if 'date' == rand_type:
            start = pd.Timestamp(self.cfg.get('start') or '19000101')
            end = pd.Timestamp(self.cfg.get('end') or '21991231')
            business_days = self.cfg.get('businessDay') is True
            timestamps = self.cfg.get('timestamps') is True
            if timestamps:
                code = (
                    'def pp(start, end, n):\n'
                    '\tstart_u = start.value // 10 ** 9\n'
                    '\tend_u = end.value // 10 ** 9\n'
                    '\treturn pd.DatetimeIndex(\n'
                    "\t\t(10 ** 9 * np.random.randint(start_u, end_u, n, dtype=np.int64)).view('M8[ns]')\n"
                    ')\n\n'
                    "df.loc[:, '{name}'] = pd.Series(\n"
                    "\tpp(pd.Timestamp('{start}'), pd.Timestamp('{end}'), len(df)), index=df.index\n"
                    ')'
                ).format(name=self.name, start=start.strftime('%Y%m%d'), end=end.strftime('%Y%m%d'))
            else:
                freq = ", freq='B'" if business_days else ''
                code = (
                    "dates = pd.date_range('{start}', '{end}'{freq})\n"
                    'dates = [dates[i] for i in np.random.randint(0, len(dates) - 1, size=len(data))]\n'
                    "df.loc[:, '{name}'] = pd.Series(dates, index=data.index)"
                ).format(name=self.name, start=start.strftime('%Y%m%d'), end=end.strftime('%Y%m%d'), freq=freq)
            return code
        if 'choice' == rand_type:
            choices = self.cfg.get('choices') or 'a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z'
            choices = choices.split(',')
            return "df.loc[:, '{name}'] = pd.Series(np.random.choice({choices}, len(df)), index=df.index)".format(
                choices="', '".join(choices), name=self.name
            )

        if 'int' == rand_type:
            low = self.cfg.get('low', 0)
            high = self.cfg.get('high', 100)
            return (
                'import numpy as np\n\n'
                "df.loc[:, '{name}'] = pd.Series(np.random.randint({low}, high={high}, size=len(df)), "
                'index=df.index)'
            ).format(name=self.name, low=low, high=high)

        low = self.cfg.get('low', 0)
        high = self.cfg.get('high', 1)
        return (
            'import numpy as np\n\n'
            "df.loc[:, '{name}'] = pd.Series(np.random.uniform({low}, {high}, len(df)), index=df.index)"
        ).format(low=low, high=high, name=self.name)


class TypeConversionColumnBuilder(object):

    def __init__(self, name, cfg):
        self.name = name
        self.cfg = cfg

    def build_column(self, data):
        col, from_type, to_type = (self.cfg.get(p) for p in ['col', 'from', 'to'])
        s = data[col]
        classifier = classify_type(from_type)
        if classifier == 'S':  # col can be (str or category) -> date, int, float, bool, category
            if to_type == 'date':
                date_kwargs = {}
                if self.cfg.get('fmt'):
                    date_kwargs['format'] = self.cfg['fmt']
                else:
                    date_kwargs['infer_datetime_format'] = True
                return pd.Series(pd.to_datetime(s, **date_kwargs), name=self.name, index=s.index)
            elif to_type == 'int':
                return pd.Series(s.astype('float').astype('int'), name=self.name, index=s.index)
            else:
                return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        elif classifier == 'I':  # date, float, category, str, bool
            if to_type == 'date':
                unit = self.cfg.get('unit') or 'D'
                if unit == 'YYYYMMDD':
                    return pd.Series(s.astype(str).apply(pd.Timestamp), name=self.name, index=s.index)
                return pd.Series(pd.to_datetime(s, unit=unit), name=self.name, index=s.index)
            return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        elif classifier == 'F':  # str, int
            return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        elif classifier == 'D':  # str, int
            if to_type == 'int':
                unit = self.cfg.get('unit')
                if unit == 'YYYYMMDD':
                    return pd.Series(s.dt.strftime('%Y%m%d').astype(int), name=self.name, index=s.index)
                return pd.Series(s.apply(lambda x: time.mktime(x.timetuple())).astype(int))
            return pd.Series(s.dt.strftime(self.cfg.get('fmt') or '%Y%m%d'), name=self.name, index=s.index)
        elif classifier == 'B':
            return pd.Series(s.astype(to_type), name=self.name, index=s.index)
        raise NotImplementedError('data type conversion not supported for dtype: {}'.format(from_type))

    def build_inner_code(self):
        col, from_type, to_type = (self.cfg.get(p) for p in ['col', 'from', 'to'])
        s = "df['{col}']".format(col=col)
        classifier = classify_type(from_type)
        if classifier == 'S':  # date, int, float, bool, category
            if to_type == 'date':
                if self.cfg.get('fmt'):
                    date_kwargs = "format='{}'".format(self.cfg['fmt'])
                else:
                    date_kwargs = 'infer_datetime_format=True'
                code = "pd.Series(pd.to_datetime({s}, {kwargs}), name='{name}', index={s}.index)"
                return code.format(s=s, name=self.name, kwargs=date_kwargs)
            elif to_type == 'int':
                return "pd.Series({s}.astype('float').astype('int'), name='{name}', index={s}.index)".format(
                    s=s, name=self.name
                )
            else:
                return "pd.Series({s}.astype({to_type}), name='{name}', index={s}.index)".format(
                    s=s, to_type=to_type, name=self.name
                )
        elif classifier == 'I':  # date, float, category, str, bool
            if to_type == 'date':
                unit = self.cfg.get('unit') or 'D'
                if unit == 'YYYYMMDD':
                    return "pd.Series({s}.astype(str).apply(pd.Timestamp), name='{name}', index={s}.index)".format(
                        s=s, name=self.name,
                    )
                return "pd.Series(pd.to_datetime({s}, unit='{unit}'), name='{name}', index={s}.index)".format(
                    s=s, name=self.name, unit=unit
                )
            return "pd.Series({s}.astype('{to_type}'), name='{name}', index={s}.index)".format(
                s=s, to_type=to_type, name=self.name
            )
        elif classifier == 'F':  # str, int
            return "pd.Series(s.astype('{to_type}'), name='{name}', index={s}.index)".format(
                s=s, to_type=to_type, name=self.name
            )
        elif classifier == 'D':  # str, int
            if to_type == 'int':
                unit = self.cfg.get('unit') or 'D'
                if unit == 'YYYYMMDD':
                    return "pd.Series({s}.dt.strftime('%Y%m%d').astype(int), name='{name}', index={s}.index)".format(
                        s=s, name=self.name
                    )
                return (
                    "pd.Series(\n"
                    "\t{s}.apply(lambda x: time.mktime(x.timetuple())).astype(int), \n"
                    "name='{name}', index={s}.index\n"
                    ")"
                ).format(s=s, name=self.name)
            return "pd.Series({s}.dt.strftime('{fmt}'), name='{name}', index={s}.index)".format(
                fmt=self.cfg.get('fmt') or '%Y%m%d', s=s, name=self.name
            )
        elif classifier == 'B':
            return "pd.Series(s.astype('{to_type}'), name='{name}', index={s}.index)".format(
                s=s, to_type=to_type, name=self.name
            )
        raise NotImplementedError('data type conversion not supported for dtype: {}'.format(from_type))

    def build_code(self):
        code = self.build_inner_code()
        return "df.loc[:, '{name}'] = {code}".format(name=self.name, code=code)



import json

import dtale.global_state as global_state
from dtale.utils import classify_type, find_dtype, make_list


class ColumnFilter(object):

    def __init__(self, data_id, column, cfg):
        self.data_id = data_id
        self.column = column
        s = global_state.get_data(data_id)[column]
        dtype = find_dtype(s)
        self.classification = classify_type(dtype)
        self.cfg = cfg
        if self.cfg is not None:
            self.cfg = json.loads(self.cfg)
        if self.cfg['type'] == 'string':
            self.builder = StringFilter(column, self.classification, self.cfg)
        if self.cfg['type'] in ['int', 'float']:
            self.builder = NumericFilter(column, self.classification, self.cfg)
        if self.cfg['type'] == 'date':
            self.builder = DateFilter(column, self.classification, self.cfg)
        if self.cfg['type'] == 'outliers':
            self.builder = OutlierFilter(column, self.classification, self.cfg)

    def save_filter(self):
        curr_settings = global_state.get_settings(self.data_id)
        filters_key = '{}Filters'.format('outlier' if self.cfg['type'] == 'outliers' else 'column')
        curr_filters = curr_settings.get(filters_key) or {}
        fltr = self.builder.build_filter()
        if fltr is None:
            curr_filters.pop(self.column, None)
        else:
            curr_filters[self.column] = fltr
        curr_settings[filters_key] = curr_filters
        global_state.set_settings(self.data_id, curr_settings)
        return curr_filters


class OutlierFilter(object):
    def __init__(self, column, classification, cfg):
        self.column = column
        self.classification = classification
        self.cfg = cfg

    def build_filter(self):
        if self.cfg.get("query") is None:
            return None
        return self.cfg


class MissingFilter(object):
    def __init__(self, column, classification, cfg):
        self.column = column
        self.classification = classification
        self.cfg = cfg

    def handle_missing(self, fltr):
        if self.cfg is None or not self.cfg.get('missing', False):
            return fltr
        return {'missing': True, 'query': '{col} != {col}'.format(col=self.column)}


class StringFilter(MissingFilter):

    def __init__(self, column, classification, cfg):
        super(StringFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None or not len(self.cfg.get('value', [])):
            return super(StringFilter, self).handle_missing(None)

        state = self.cfg.get('value', [])
        operand = self.cfg.get('operand', '=')
        fltr = dict(value=state)
        if len(state) == 1:
            val_str = ("'{}'" if self.classification == 'S' else '{}').format(state[0])
            fltr['query'] = "{} {} {}".format(self.column, '==' if operand == '=' else '!=', val_str)
        else:
            val_str = ("'{}'".format("', '".join(state)) if self.classification == 'S' else ','.join(state))
            fltr['query'] = "{} {} ({})".format(self.column, 'in' if operand == '=' else 'not in', val_str)
        return super(StringFilter, self).handle_missing(fltr)


class NumericFilter(MissingFilter):

    def __init__(self, column, classification, cfg):
        super(NumericFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None:
            return super(NumericFilter, self).handle_missing(None)
        cfg_val, cfg_operand, cfg_min, cfg_max = (self.cfg.get(p) for p in ['value', 'operand', 'min', 'max'])

        if cfg_operand in ['=', 'ne']:
            state = make_list(cfg_val or [])
            if not len(state):
                return super(NumericFilter, self).handle_missing(None)
            fltr = dict(value=cfg_val, operand=cfg_operand)
            if len(state) == 1:
                fltr['query'] = "{} {} {}".format(self.column, '==' if cfg_operand == '=' else '!=', state[0])
            else:
                fltr['query'] = "{} {} ({})".format(
                    self.column, 'in' if cfg_operand == '=' else 'not in', ", ".join(state)
                )
            return super(NumericFilter, self).handle_missing(fltr)
        if cfg_operand in ['<', '>', '<=', '>=']:
            if cfg_val is None:
                return super(NumericFilter, self).handle_missing(None)
            fltr = dict(value=cfg_val, operand=cfg_operand, query='{} {} {}'.format(self.column, cfg_operand, cfg_val))
            return super(NumericFilter, self).handle_missing(fltr)
        if cfg_operand in ['[]', '()']:
            fltr = dict(operand=cfg_operand)
            queries = []
            if cfg_min is not None:
                fltr['min'] = cfg_min
                queries.append('{} >{} {}'.format(self.column, '=' if cfg_operand == '[]' else '', cfg_min))
            if cfg_max is not None:
                fltr['max'] = cfg_max
                queries.append('{} <{} {}'.format(self.column, '=' if cfg_operand == '[]' else '', cfg_max))
            if len(queries) == 2 and cfg_max == cfg_min:
                queries = ["{} == {}".format(self.column, cfg_max)]
            if not len(queries):
                return super(NumericFilter, self).handle_missing(None)
            fltr['query'] = ' and '.join(queries)
            return super(NumericFilter, self).handle_missing(fltr)
        return super(NumericFilter, self).handle_missing(None)


class DateFilter(MissingFilter):

    def __init__(self, column, classification, cfg):
        super(DateFilter, self).__init__(column, classification, cfg)

    def build_filter(self):
        if self.cfg is None:
            return super(DateFilter, self).handle_missing(None)

        start, end = (self.cfg.get(p) for p in ['start', 'end'])
        fltr = dict(start=start, end=end)
        queries = []
        if start:
            queries.append("{} >= '{}'".format(self.column, start))
        if end:
            queries.append("{} <= '{}'".format(self.column, end))
        if len(queries) == 2 and start == end:
            queries = ["{} == '{}'".format(self.column, start)]
        if not len(queries):
            return super(DateFilter, self).handle_missing(None)
        fltr['query'] = ' and '.join(queries)
        return super(DateFilter, self).handle_missing(fltr)



from flask import Blueprint

dtale = Blueprint('dtale', __name__, url_prefix='/dtale')

# flake8: NOQA
from dtale.app import show, get_instance, instances, offline_chart  # isort:skip
from dtale.cli.loaders import LOADERS  # isort:skip

for loader_name, loader in LOADERS.items():
    if hasattr(loader, 'show_loader'):
        globals()['show_{}'.format(loader_name)] = loader.show_loader



import json

import dash_core_components as dcc
import dash_daq as daq
import dash_html_components as html
import plotly
from pkg_resources import parse_version

from dtale.charts.utils import YAXIS_CHARTS, ZAXIS_CHARTS, find_group_vals
from dtale.utils import (ChartBuildingError, classify_type, dict_merge,
                         flatten_lists, get_dtypes, inner_build_query,
                         make_list)


def test_plotly_version(version_num):
    if 'unknown' in plotly.__version__:
        return True
    return parse_version(plotly.__version__) >= parse_version(version_num)


def base_layout(github_fork, **kwargs):
    """
    Base layout to be returned by :meth:`dtale.dash_application.views.DtaleDash.interpolate_index`

    :param github_fork: `True` if "Fork me on Github" banner should be displayed, `False` otherwise
    :type github_fork: bool
    :param kwargs: Optional keyword arguments to be passed to 'dash.Dash.interplolate_index'
    :type kwargs: dict
    :return: HTML
    :rtype: str
    """
    back_to_data_padding, github_fork_html = ('', '')
    if github_fork:
        back_to_data_padding = 'padding-right: 125px'
        github_fork_html = '''
            <span id="forkongithub">
                <a href="https://github.com/man-group/dtale">Fork me on GitHub</a>
            </span>
        '''
    return '''
        <!DOCTYPE html>
        <html>
            <head>
                {metas}
                <title>D-Tale Charts</title>
                <link rel="shortcut icon" href="../../images/favicon.png">
                {css}
            </head>
            <body>
                {github_fork_html}
                <div class="container-fluid charts">
                    <div class="row" style="margin: 0">
                        <div class="col-auto">
                            <header>
                                <span class="title-font">D-TALE</span>
                                <span style="font-size: 16px" class="pl-4">Charts</span>
                            </header>
                        </div>
                        <div class="col"></div>
                        <div class="col-auto mt-4" style="{back_to_data_padding}">
                            <a href="#" onclick="javascript:backToData()">
                                <i class="fas fa-th mr-4"></i>
                                <span>Back To Data</span>
                            </a>
                        </div>
                    </div>
                    {app_entry}
                </div>
                <footer>
                    {config}
                    {scripts}
                    <script type="text/javascript">
                        const pathSegs = window.location.pathname.split('/');
                        const dataId = pathSegs[pathSegs.length - 1];
                        const backToData = () => window.open('/dtale/main/' + dataId);
                        const goToLegacy = () => location.replace('/dtale/popup/charts/' + dataId);
                    </script>
                    {renderer}
                    {css}
                </footer>
            </body>
        </html>
    '''.format(
        metas=kwargs['metas'],
        css=kwargs['css'],
        app_entry=kwargs['app_entry'],
        config=kwargs['config'],
        scripts=kwargs['scripts'],
        renderer=kwargs['renderer'],
        back_to_data_padding=back_to_data_padding,
        github_fork_html=github_fork_html
    )


def build_input(label, input, className='col-auto', label_class='input-group-addon', **kwargs):
    """
    Helper function to build a standard label/input component in dash.

    :param label: name of the input you are displaying
    :type label: str
    :param input: dash component for storing state
    :param className: style class to be applied to encapsulating div
    :type className: str
    :param kwargs: Optional keyword arguments to be applied to encapsulating div (style, title, id...)
    :type kwargs: dict
    :return: dash components for label/input
    :rtype: :dash:`dash_html_components.Div <dash-html-components/div>`
    """
    return html.Div(
        [
            html.Div(
                [html.Span(label, className=label_class), input],
                className='input-group mr-3',
            )
        ],
        className=className, **kwargs
    )


def build_tab(label, value, additional_style=None, **kwargs):
    """
    Builds a :dash:`dash_core_components.Tab <dash-core-components/tab>` with standard styling settings.
    """
    base_style = {'borderBottom': '1px solid #d6d6d6', 'padding': '6px'}
    return dcc.Tab(
        label=label,
        value=value,
        style=dict_merge(base_style, {'fontWeight': 'bold'}, additional_style or {}),
        disabled_style=dict_merge(
            base_style,
            {'fontWeight': 'bold', 'backgroundColor': 'LightGray', 'color': 'black', 'cursor': 'not-allowed'},
            additional_style or {}
        ),
        selected_style=dict_merge(
            base_style,
            {'borderTop': '1px solid #d6d6d6', 'backgroundColor': '#2a91d1', 'color': 'white'},
            additional_style or {}
        ), **kwargs)


def build_option(value, label=None):
    """
    Returns value/label inputs in a dictionary for use in
    :dash:`dash_core_components.Dropdown <dash-core-components/Dropdown>`
    """
    return {'label': label or value, 'value': value}


CHARTS = [
    dict(value='line'), dict(value='bar'), dict(value='scatter'), dict(value='pie'), dict(value='wordcloud'),
    dict(value='heatmap'), dict(value='3d_scatter', label='3D Scatter'), dict(value='surface'),
    dict(value='maps', label='Maps')
]
CHART_INPUT_SETTINGS = {
    'line': dict(x=dict(type='single'), y=dict(type='multi'), z=dict(display=False),
                 group=dict(display=True, type='single')),
    'bar': dict(x=dict(type='single'), y=dict(type='multi'), z=dict(display=False),
                group=dict(display=True, type='single')),
    'scatter': dict(x=dict(type='single'), y=dict(type='multi'), z=dict(display=False),
                    group=dict(display=True, type='single')),
    'pie': dict(x=dict(type='single'), y=dict(type='multi'), z=dict(display=False),
                group=dict(display=True, type='single')),
    'wordcloud': dict(x=dict(type='single'), y=dict(type='multi'), z=dict(display=False),
                      group=dict(display=True, type='single')),
    'heatmap': dict(x=dict(type='single'), y=dict(type='single'), z=dict(display=True, type='single'),
                    group=dict(display=False)),
    '3d_scatter': dict(x=dict(type='single'), y=dict(type='single'), z=dict(display=True, type='single'),
                       group=dict(display=True)),
    'surface': dict(x=dict(type='single'), y=dict(type='single'), z=dict(display=True, type='single'),
                    group=dict(display=False)),
    'maps': dict(x=dict(display=False), y=dict(display=False), z=dict(display=False), group=dict(display=False),
                 map_group=dict(display=True)),
}
AGGS = dict(
    raw='No Aggregation', count='Count', nunique='Unique Count', sum='Sum', mean='Mean', rolling='Rolling',
    corr='Correlation', first='First', last='Last', median='Median', min='Minimum', max='Maximum',
    std='Standard Deviation', var='Variance', mad='Mean Absolute Deviation', prod='Product of All Items',
    pctct='Percentage Count', pctsum='Percentage Sum'
)
FREQS = ['H', 'H2', 'WD', 'D', 'W', 'M', 'Q', 'Y']
FREQ_LABELS = dict(H='Hourly', H2='Hour', WD='Weekday', W='Weekly', M='Monthly', Q='Quarterly', Y='Yearly')

MAP_TYPES = [dict(value='choropleth'), dict(value='scattergeo', label='ScatterGeo')]
SCOPES = ['world', 'usa', 'europe', 'asia', 'africa', 'north america', 'south america']
PROJECTIONS = ['equirectangular', 'mercator', 'orthographic', 'natural earth', 'kavrayskiy7', 'miller', 'robinson',
               'eckert4', 'azimuthal equal area', 'azimuthal equidistant', 'conic equal area', 'conic conformal',
               'conic equidistant', 'gnomonic', 'stereographic', 'mollweide', 'hammer', 'transverse mercator',
               'albers usa', 'winkel tripel', 'aitoff', 'sinusoidal']


def build_img_src(proj, img_type='projections'):
    return '/images/{}/{}.png'.format(img_type, '_'.join(proj.split(' ')))


def build_proj_hover_children(proj):
    if proj is None:
        return None
    return [
        html.I(className='ico-help-outline', style=dict(color='white')),
        html.Div(
            [html.Div(proj), html.Img(src=build_img_src(proj))],
            className='hoverable__content',
            style=dict(width='auto')
        )
    ]


def build_proj_hover(proj):
    return html.Span(
        [
            'Projection',
            html.Div(
                build_proj_hover_children(proj),
                className='ml-3 hoverable',
                style=dict(display='none') if proj is None else dict(borderBottom='none'),
                id='proj-hover'
            )
        ],
        className='input-group-addon'
    )


LOC_MODE_INFO = {
    'ISO-3': dict(
        url=html.A('ISO-3', href='https://en.wikipedia.org/wiki/ISO_3166-1_alpha-3', target='_blank'),
        examples=['USA (United States)', 'CAN (Canada)', 'GBR (United Kingdom)']
    ),
    'USA-states': dict(
        url=html.A('USA-states (use ISO)', href='https://en.wikipedia.org/wiki/List_of_U.S._state_abbreviations',
                   target='_blank'),
        examples=['NY (New York)', 'CA (California)', 'MA (Massachusetts)']
    ),
    'country names': dict(
        url=html.A('country names (case-insensitive)', href='https://simple.wikipedia.org/wiki/List_of_countries',
                   target='_blank'),
        examples=['United States', 'United Kingdom', 'Germany']
    )
}


def build_loc_mode_hover_children(loc_mode):
    if loc_mode is None:
        return None
    loc_mode_cfg = LOC_MODE_INFO[loc_mode]
    return [
        html.I(className='ico-help-outline', style=dict(color='white')),
        html.Div(
            [
                html.Div([html.Span('View', className='mr-3'), loc_mode_cfg['url']]),
                html.Ul(
                    [html.Li(e, style={'listStyleType': 'disc'}, className='mb-3') for e in loc_mode_cfg['examples']],
                    className='pt-3 mb-0'
                )
            ],
            className='hoverable__content build-code',
            style=dict(width='auto', whiteSpace='nowrap', left='-2em', top='auto')
        )
    ]


def build_loc_mode_hover(loc_mode):
    return html.Span(
        [
            html.Span('Location Mode', style=dict(whiteSpace='pre-line')),
            html.Div(
                build_loc_mode_hover_children(loc_mode),
                className='ml-3 hoverable',
                style=dict(display='none') if loc_mode is None else dict(borderBottom='none'),
                id='loc-mode-hover'
            )
        ],
        className='input-group-addon pt-1 pb-0'
    )


COLORSCALES = ['Blackbody', 'Bluered', 'Blues', 'Earth', 'Electric', 'Greens', 'Greys', 'Hot', 'Jet', 'Picnic',
               'Portland', 'Rainbow', 'RdBu', 'Reds', 'Viridis', 'YlGnBu', 'YlOrRd']

ANIMATION_CHARTS = ['line']
ANIMATE_BY_CHARTS = ['bar', '3d_scatter', 'maps']


def show_input_handler(chart_type):
    settings = CHART_INPUT_SETTINGS.get(chart_type or 'line') or {}

    def _show_input(input_id, input_type='single'):
        cfg = settings.get(input_id, {})
        return cfg.get('display', True) and cfg.get('type', 'single') == input_type
    return _show_input


def update_label_for_freq(val):
    """
    Formats sub-values contained within 'val' to display date frequencies if included.
        - (val=['a', 'b', 'c']) => 'a, b, c'
        - (val=['a|H', 'b', 'c']) => 'a (Hour), b, c'
    """

    def _freq_handler(sub_val):
        for freq in FREQS:
            if sub_val.endswith('|{}'.format(freq)):
                col, freq = sub_val.split('|')
                if freq in FREQ_LABELS:
                    return '{} ({})'.format(col, FREQ_LABELS[freq])
                return col
        return sub_val
    return ', '.join([_freq_handler(sub_val) for sub_val in make_list(val)])


def build_error(error, tb):
    """
    Returns error/traceback information in standard component with styling

    :param error: execption message
    :type error: str
    :param tb: tracebackF
    :type tb: str
    :return: error component
    :rtype: :dash:`dash_html_components.Div <dash-html-components/div>`
    """
    if isinstance(error, ChartBuildingError):
        if error.details:
            tb = error.details
        error = error.error
    return html.Div([
        html.I(className='ico-error'), html.Span(str(error)), html.Div(html.Pre(str(tb)), className='traceback')
    ], className='dtale-alert alert alert-danger')


def build_cols(cols, dtypes):
    """
    Helper function to add additional column entries for columns of type datetime so that users can make use of
    different frequencies of dates.  For example, hour, weekday, month, quarter, year

    :param cols: columns in dataframe
    :type cols: list of strings
    :param dtypes: datatypes of columns in dataframe
    :type dtypes: dict
    :return: generator or columns + any additional (datetime column + frequency) options
    """
    for c in cols:
        if classify_type(dtypes[c]) == 'D':
            for freq in FREQS:
                if freq in FREQ_LABELS:
                    yield '{}|{}'.format(c, freq), '{} ({})'.format(c, FREQ_LABELS[freq])
                else:
                    yield c, c
        else:
            yield c, c


def build_selections(*args):
    """
    simple helper function to build a single level list of values based on variable number of inputs which could be
    equal to None.
    """
    return flatten_lists([[] if a is None else make_list(a) for a in args])


def build_input_options(df, **inputs):
    """
    Builds dropdown options for (X, Y, Z, Group, Barsort & Y-Axis Ranges) with filtering based on currently selected
    values for the following inputs: x, y, z, group.
    """
    chart_type, x, y, z, group = (inputs.get(p) for p in ['chart_type', 'x', 'y', 'z', 'group'])
    col_opts = list(build_cols(df.columns, get_dtypes(df)))
    group_val, z_val = (None, z) if chart_type in ZAXIS_CHARTS else (group, None)
    x_options = [build_option(c, l) for c, l in col_opts if c not in build_selections(y, z_val, group_val)]
    y_filter = build_selections(x, group_val, z_val)
    y_multi_options = [build_option(c, l) for c, l in col_opts if c not in y_filter]
    y_single_options = [build_option(c, l) for c, l in col_opts if c not in y_filter]
    z_options = [build_option(c) for c in df.columns if c not in build_selections(x, y, group_val)]
    group_options = [build_option(c, l) for c, l in col_opts if c not in build_selections(x, y, z_val)]
    barsort_options = [build_option(o) for o in build_selections(x, y)]
    yaxis_options = [build_option(y2) for y2 in y or []]

    return x_options, y_multi_options, y_single_options, z_options, group_options, barsort_options, yaxis_options


def build_map_options(df, type='choropleth', loc=None, lat=None, lon=None, map_val=None):
    dtypes = get_dtypes(df)
    cols = sorted(dtypes.keys())
    float_cols, str_cols, num_cols = [], [], []
    for c in cols:
        dtype = dtypes[c]
        classification = classify_type(dtype)
        if classification == 'S':
            str_cols.append(c)
            continue
        if classification in ['F', 'I']:
            num_cols.append(c)
            if classification == 'F':
                float_cols.append(c)

    lat_options = [build_option(c) for c in float_cols if c not in build_selections(lon, map_val)]
    lon_options = [build_option(c) for c in float_cols if c not in build_selections(lat, map_val)]
    loc_options = [build_option(c) for c in str_cols if c not in build_selections(map_val)]

    if type == 'choropleth':
        val_options = [build_option(c) for c in num_cols if c not in build_selections(loc)]
    else:
        val_options = [build_option(c) for c in num_cols if c not in build_selections(lon, lat)]
    return loc_options, lat_options, lon_options, val_options


def bar_input_style(**inputs):
    """
    Sets display CSS property for bar chart inputs
    """
    chart_type, group_col = (inputs.get(p) for p in ['chart_type', 'group'])
    show_bar = chart_type == 'bar'
    show_barsort = show_bar and group_col is None
    return dict(display='block' if show_bar else 'none'), dict(display='block' if show_barsort else 'none')


def colorscale_input_style(**inputs):
    return dict(display='block' if inputs.get('chart_type') in ['heatmap', 'maps'] else 'none')


def animate_styles(df, **inputs):
    chart_type, agg, cpg = (inputs.get(p) for p in ['chart_type', 'agg', 'cpg'])
    opts = []
    if cpg or agg in ['pctsum', 'pctct']:
        return dict(display='none'), dict(display='none'), opts
    if chart_type in ANIMATION_CHARTS:
        return dict(display='block'), dict(display='none'), opts
    if chart_type in ANIMATE_BY_CHARTS:
        opts = [build_option(v, l) for v, l in build_cols(df.columns, get_dtypes(df))]
    if len(opts):
        return dict(display='none'), dict(display='block'), opts
    return dict(display='none'), dict(display='none'), []


def show_chart_per_group(**inputs):
    """
    Boolean function to determine whether "Chart Per Group" toggle should be displayed or not
    """
    [chart_type, group] = [inputs.get(p) for p in ['chart_type', 'group']]
    invalid_type = chart_type in ['pie', 'wordcloud', 'maps']
    return show_input_handler(chart_type)('group') and len(group or []) and not invalid_type


def show_yaxis_ranges(**inputs):
    """
    Boolean function to determine whether "Y-Axis Range" inputs should be displayed or not
    """
    chart_type, y = (inputs.get(p) for p in ['chart_type', 'y'])
    return chart_type in YAXIS_CHARTS and len(y or [])


def get_yaxis_type_tabs(y):
    tabs = [
        build_tab('Default', 'default', {'padding': '2px', 'minWidth': '4em'}),
        build_tab('Single', 'single', {'padding': '2px', 'minWidth': '4em'}),
    ]
    if len(y) <= 1:
        return tabs
    return tabs + [build_tab('Multi', 'multi', {'padding': '2px', 'minWidth': '4em'})]


def build_group_val_options(df, group_cols):
    group_vals = find_group_vals(df, group_cols)
    return [
        build_option(json.dumps(gv), '|'.join([str(gv.get(p, 'NaN')) for p in group_cols]))
        for gv in group_vals
    ]


def build_map_type_tabs(map_type):
    def _build_hoverable():
        for t in MAP_TYPES:
            yield html.Div([
                html.Span(t.get('label', t['value'].capitalize())),
                html.Img(src=build_img_src(t['value'], img_type='map_type'))
            ], className='col-md-6')

    return html.Div(
        [dcc.Tabs(
            id='map-type-tabs',
            value=map_type or 'choropleth',
            children=[build_tab(t.get('label', t['value'].capitalize()), t['value']) for t in MAP_TYPES],
            style=dict(height='36px'),
        ), html.Div(
            html.Div(list(_build_hoverable()), className='row'),
            className='hoverable__content map-types'
        )],
        style=dict(paddingLeft=15, borderBottom='none'), className="pr-5 hoverable"
    )


def main_inputs_and_group_val_display(inputs):
    chart_type = inputs.get('chart_type')
    show_group = show_input_handler(inputs.get('chart_type', 'line'))('group')
    if chart_type == 'maps' and not len(make_list(inputs.get('map_group'))):
        return dict(display='none'), 'col-md-12'
    elif show_group and not len(make_list(inputs.get('group'))):
        return dict(display='none'), 'col-md-12'
    return dict(display='block'), 'col-md-8'


def charts_layout(df, settings, **inputs):
    """
    Builds main dash inputs with dropdown options populated with the columns of the dataframe associated with the
    page. Inputs included are: chart tabs, query, x, y, z, group, aggregation, rolling window/computation,
    chart per group toggle, bar sort, bar mode, y-axis range editors

    :param df: dataframe to drive the charts built on page
    :type df: :class:`pandas:pandas.DataFrame`
    :param settings: global settings associated with this dataframe (contains properties like "query")
    :type param: dict
    :return: dash markup
    """
    chart_type, x, y, z, group, agg = (inputs.get(p) for p in ['chart_type', 'x', 'y', 'z', 'group', 'agg'])
    y = y or []
    show_input = show_input_handler(chart_type)
    show_cpg = show_chart_per_group(**inputs)
    show_yaxis = show_yaxis_ranges(**inputs)
    bar_style, barsort_input_style = bar_input_style(**inputs)
    animate_style, animate_by_style, animate_opts = animate_styles(df, **inputs)

    options = build_input_options(df, **inputs)
    x_options, y_multi_options, y_single_options, z_options, group_options, barsort_options, yaxis_options = options
    query_placeholder = (
        "Enter pandas query (ex: col1 == 1)"
    )
    query_value = inputs.get('query') or inner_build_query(settings, settings.get('query'))
    query_label = html.Div([
        html.Span('Query'),
        html.A(html.I(className='fa fa-info-circle ml-4'),
               href='https://pandas.pydata.org/pandas-docs/stable/user_guide/indexing.html#indexing-query',
               target='_blank', style={'color': 'white'})
    ], className='input-group-addon', style={'minWidth': '7em'})
    yaxis_type = (inputs.get('yaxis') or {}).get('type') or 'default'
    yaxis_type_style = {'borderRadius': '0 0.25rem 0.25rem 0'} if yaxis_type == 'default' else None
    show_map = chart_type == 'maps'
    map_props = ['map_type', 'loc_mode', 'loc', 'lat', 'lon', 'map_val']
    map_type, loc_mode, loc, lat, lon, map_val = (inputs.get(p) for p in map_props)
    map_scope, proj = (inputs.get(p) for p in ['scope', 'proj'])
    loc_options, lat_options, lon_options, map_val_options = build_map_options(df, type=map_type, loc=loc, lat=lat,
                                                                               lon=lon, map_val=map_val)
    cscale_style = colorscale_input_style(**inputs)
    default_cscale = 'Greens' if chart_type == 'heatmap' else 'Reds'

    group_val_style, main_input_class = main_inputs_and_group_val_display(inputs)
    group_val = [json.dumps(gv) for gv in inputs.get('group_val') or []]

    def show_style(show):
        return {'display': 'block' if show else 'none'}

    def show_map_style(show):
        return {} if show else {'display': 'none'}
    return html.Div([
        dcc.Store(id='query-data', data=inputs.get('query')),
        dcc.Store(id='input-data', data={k: v for k, v in inputs.items() if k not in ['cpg', 'barmode', 'barsort']}),
        dcc.Store(id='chart-input-data', data={k: v for k, v in inputs.items() if k in ['cpg', 'barmode', 'barsort']}),
        dcc.Store(
            id='map-input-data',
            data={k: v for k, v in inputs.items() if k in ['map_type', 'map_code', 'lat', 'lon', 'map_val', 'scope',
                                                           'proj']}
        ),
        dcc.Store(id='range-data'),
        dcc.Store(id='yaxis-data', data=inputs.get('yaxis')),
        dcc.Store(id='last-chart-input-data', data=inputs),
        dcc.Input(id='chart-code', type='hidden'),
        html.Div(html.Div(dcc.Tabs(
            id='chart-tabs',
            value=chart_type or 'line',
            children=[build_tab(t.get('label', t['value'].capitalize()), t['value']) for t in CHARTS],
            style=dict(height='36px')
        ), className='col-md-12'), className='row pt-3 pb-3 charts-filters'),
        html.Div(html.Div([
            html.Div([
                query_label, dcc.Input(
                    id='query-input', type='text', placeholder=query_placeholder, className='form-control',
                    value=query_value, style={'lineHeight': 'inherit'})
            ], className='input-group mr-3')],
            className='col'
        ), className='row pt-3 pb-3 charts-filters'),
        html.Div(
            [html.Div([
                html.Div(
                    [
                        build_input(
                            [html.Div('X'), html.Small('(Agg By)')],
                            dcc.Dropdown(
                                id='x-dropdown',
                                options=x_options,
                                placeholder='Select a column',
                                value=x,
                                style=dict(width='inherit'),
                            ),
                            label_class='input-group-addon d-block pt-1 pb-0'
                        ),
                        build_input(
                            'Y',
                            dcc.Dropdown(
                                id='y-multi-dropdown',
                                options=y_multi_options,
                                multi=True,
                                placeholder='Select a column(s)',
                                style=dict(width='inherit'),
                                value=y if show_input('y', 'multi') else None
                            ),
                            className='col',
                            id='y-multi-input',
                            style=show_style(show_input('y', 'multi'))
                        ),
                        build_input(
                            [html.Div('Y'), html.Small('(Agg By)')],
                            dcc.Dropdown(
                                id='y-single-dropdown',
                                options=y_single_options,
                                placeholder='Select a column',
                                style=dict(width='inherit'),
                                value=y[0] if show_input('y') and len(y) else None
                            ),
                            className='col',
                            label_class='input-group-addon d-block pt-1 pb-0',
                            id='y-single-input',
                            style=show_style(show_input('y'))
                        ),
                        build_input('Z', dcc.Dropdown(
                            id='z-dropdown',
                            options=z_options,
                            placeholder='Select a column',
                            style=dict(width='inherit'),
                            value=z
                        ), className='col', id='z-input', style=show_style(show_input('z'))),
                        build_input(
                            'Group',
                            dcc.Dropdown(
                                id='group-dropdown',
                                options=group_options,
                                multi=True,
                                placeholder='Select a group(s)',
                                value=group,
                                style=dict(width='inherit'),
                            ),
                            className='col',
                            id='group-input',
                            style=show_style(show_input('group'))
                        )
                    ],
                    id='non-map-inputs', style={} if not show_map else {'display': 'none'},
                    className='row p-0 charts-filters'
                ),
                html.Div(
                    [
                        build_map_type_tabs(map_type),
                        html.Div(
                            [
                                html.Div(
                                    [
                                        build_loc_mode_hover(loc_mode),
                                        dcc.Dropdown(
                                            id='map-loc-mode-dropdown',
                                            options=[build_option(v) for v in ["ISO-3", "USA-states", "country names"]],
                                            style=dict(width='inherit'),
                                            value=loc_mode
                                        )
                                    ],
                                    className='input-group mr-3',
                                )
                            ],
                            id='map-loc-mode-input', style=show_map_style(map_type == 'choropleth'),
                            className='col-auto'
                        ),
                        build_input(
                            [html.Div('Locations'), html.Small('(Agg By)')],
                            dcc.Dropdown(
                                id='map-loc-dropdown',
                                options=loc_options,
                                placeholder='Select a column',
                                value=loc,
                                style=dict(width='inherit'),
                            ),
                            id='map-loc-input',
                            label_class='input-group-addon d-block pt-1 pb-0',
                            style=show_map_style(map_type == 'choropleth')
                        ),
                        build_input(
                            [html.Div('Lat'), html.Small('(Agg By)')],
                            dcc.Dropdown(
                                id='map-lat-dropdown',
                                options=lat_options,
                                placeholder='Select a column',
                                value=lat,
                                style=dict(width='inherit'),
                            ),
                            id='map-lat-input',
                            label_class='input-group-addon d-block pt-1 pb-0',
                            style=show_map_style(map_type == 'scattergeo')
                        ),
                        build_input(
                            [html.Div('Lon'), html.Small('(Agg By)')],
                            dcc.Dropdown(
                                id='map-lon-dropdown',
                                options=lon_options,
                                placeholder='Select a column',
                                style=dict(width='inherit'),
                                value=lon
                            ),
                            id='map-lon-input',
                            label_class='input-group-addon d-block pt-1 pb-0',
                            style=show_map_style(map_type == 'scattergeo')
                        ),
                        build_input('Scope', dcc.Dropdown(
                            id='map-scope-dropdown',
                            options=[build_option(v) for v in SCOPES],
                            style=dict(width='inherit'),
                            value=map_scope or 'world'
                        ), id='map-scope-input', style=show_map_style(map_type == 'scattergeo')),
                        html.Div(
                            [
                                html.Div(
                                    [
                                        build_proj_hover(proj),
                                        dcc.Dropdown(
                                            id='map-proj-dropdown',
                                            options=[build_option(v) for v in PROJECTIONS],
                                            style=dict(width='inherit'),
                                            value=proj
                                        )
                                    ],
                                    className='input-group mr-3',
                                )
                            ],
                            id='map-proj-input', style=show_map_style(map_type == 'scattergeo'), className='col-auto'
                        ),
                        build_input('Value', dcc.Dropdown(
                            id='map-val-dropdown',
                            options=map_val_options,
                            placeholder='Select a column',
                            style=dict(width='inherit'),
                            value=map_val
                        )),
                        build_input(
                            'Group',
                            dcc.Dropdown(
                                id='map-group-dropdown',
                                options=group_options,
                                multi=True,
                                placeholder='Select a group(s)',
                                value=inputs.get('map_group'),
                                style=dict(width='inherit'),
                            ),
                            className='col',
                            id='map-group-input'
                        )
                    ],
                    id='map-inputs', className='row pt-3 pb-3 charts-filters',
                    style={} if show_map else {'display': 'none'}
                ),
                html.Div([
                    build_input('Aggregation', dcc.Dropdown(
                        id='agg-dropdown',
                        options=[build_option(v, AGGS[v]) for v in ['raw', 'count', 'nunique', 'sum', 'mean', 'rolling',
                                                                    'corr', 'first', 'last', 'median', 'min', 'max',
                                                                    'std', 'var', 'mad', 'prod', 'pctsum', 'pctct']],
                        placeholder='Select an aggregation',
                        style=dict(width='inherit'),
                        value=agg or 'raw',
                    )),
                    html.Div([
                        build_input('Window', dcc.Input(
                            id='window-input', type='number', placeholder='Enter days',
                            className='form-control text-center', style={'lineHeight': 'inherit'},
                            value=inputs.get('window')
                        )),
                        build_input('Computation', dcc.Dropdown(
                            id='rolling-comp-dropdown',
                            options=[
                                build_option('corr', 'Correlation'),
                                build_option('count', 'Count'),
                                build_option('cov', 'Covariance'),
                                build_option('kurt', 'Kurtosis'),
                                build_option('max', 'Maximum'),
                                build_option('mean', 'Mean'),
                                build_option('median', 'Median'),
                                build_option('min', 'Minimum'),
                                build_option('skew', 'Skew'),
                                build_option('std', 'Standard Deviation'),
                                build_option('sum', 'Sum'),
                                build_option('var', 'Variance'),
                            ],
                            placeholder='Select an computation',
                            style=dict(width='inherit'), value=inputs.get('rolling_comp')
                        ))
                    ], id='rolling-inputs', style=show_style(agg == 'rolling'))
                ], className='row pt-3 pb-3 charts-filters'),
                html.Div(
                    [
                        build_input('Chart Per\nGroup',
                                    html.Div(daq.BooleanSwitch(id='cpg-toggle', on=inputs.get('cpg') or False),
                                             className='toggle-wrapper'),
                                    id='cpg-input', style=show_style(show_cpg), className='col-auto'),
                        build_input('Barmode', dcc.Dropdown(
                            id='barmode-dropdown',
                            options=[
                                build_option('group', 'Group'),
                                build_option('stack', 'Stack'),
                                build_option('relative', 'Relative'),
                            ],
                            value=inputs.get('barmode') or 'group',
                            placeholder='Select a mode',
                        ), className='col-auto addon-min-width', style=bar_style, id='barmode-input'),
                        build_input('Barsort', dcc.Dropdown(
                            id='barsort-dropdown', options=barsort_options, value=inputs.get('barsort')
                        ), className='col-auto addon-min-width', style=barsort_input_style, id='barsort-input'),
                        html.Div(
                            html.Div(
                                [
                                    html.Span('Y-Axis', className='input-group-addon'),
                                    html.Div(
                                        dcc.Tabs(
                                            id='yaxis-type',
                                            value=yaxis_type,
                                            children=get_yaxis_type_tabs(y),
                                        ),
                                        id='yaxis-type-div',
                                        className='form-control col-auto pt-3',
                                        style=yaxis_type_style
                                    ),
                                    dcc.Dropdown(id='yaxis-dropdown', options=yaxis_options),
                                    html.Span('Min:', className='input-group-addon col-auto', id='yaxis-min-label'),
                                    dcc.Input(
                                        id='yaxis-min-input', type='number', className='form-control col-auto',
                                        style={'lineHeight': 'inherit'}
                                    ),
                                    html.Span('Max:', className='input-group-addon col-auto', id='yaxis-max-label'),
                                    dcc.Input(
                                        id='yaxis-max-input', type='number', className='form-control col-auto',
                                        style={'lineHeight': 'inherit'}
                                    )
                                ],
                                className='input-group', id='yaxis-min-max-options',
                            ),
                            className='col-auto addon-min-width', id='yaxis-input',
                            style=show_style(show_yaxis)
                        ),
                        build_input('Colorscale', dcc.Dropdown(
                            id='colorscale-dropdown', options=[build_option(o) for o in COLORSCALES],
                            value=inputs.get('colorscale') or default_cscale
                        ), className='col-auto addon-min-width', style=cscale_style, id='colorscale-input'),
                        build_input(
                            'Animate',
                            html.Div(daq.BooleanSwitch(id='animate-toggle', on=inputs.get('animate') or False),
                                     className='toggle-wrapper'),
                            id='animate-input',
                            style=animate_style,
                            className='col-auto'
                        ),
                        build_input('Animate By', dcc.Dropdown(
                            id='animate-by-dropdown', options=animate_opts,
                            value=inputs.get('animate_by')
                        ), className='col-auto addon-min-width', style=animate_style, id='animate-by-input'),
                    ],
                    className='row pt-3 pb-5 charts-filters', id='chart-inputs'
                )],
                id='main-inputs', className=main_input_class
            ), build_input('Group(s)', dcc.Dropdown(
                id='group-val-dropdown',
                multi=True,
                placeholder='Select a group value(s)',
                value=group_val,
                style=dict(width='inherit'),
            ), className='col-md-4 pt-3 pb-5', id='group-val-input', style=group_val_style)],
            className='row'
        ),
        dcc.Loading(html.Div(id='chart-content', style={'height': '69vh'}), type='circle'),
        dcc.Textarea(id="copy-text", style=dict(position='absolute', left='-110%'))
    ], className='charts-body')
